"""Utility functions of the cli."""
import datetime
import json
import os
import re
import random
import string
import sys
import time
import click
import requests
from terminaltables3 import AsciiTable
from textwrap import wrap

from keep import about

# Directory for Keep files
dir_path = os.path.join(os.path.expanduser('~'), '.keep')
# URL for the API
api_url = 'https://keep-cli.herokuapp.com'


def check_update(ctx, forced=False):
    """
    Check for update on pypi. Limit to 1 check per day if not forced
    """
    try:
        if ctx.update_checked and not forced:
            return
    except AttributeError:
        update_check_file = os.path.join(dir_path, 'update_check.txt')
        today = datetime.date.today().strftime("%m/%d/%Y")
        if os.path.exists(update_check_file):
            date = open(update_check_file, 'r').read()
        else:
            date = []
        if forced or today != date:
            ctx.update_checked = True
            date = today
            with open(update_check_file, 'w') as f:
                f.write(date)
            r = requests.get("https://pypi.org/pypi/keep/json").json()
            version = r['info']['version']
            curr_version = about.__version__
            if version > curr_version:
                click.secho("Keep seems to be outdated. Current version = "
                            "{}, Latest version = {}".format(curr_version, version) +
                            "\n\nPlease update with ", bold=True, fg='red')
                click.secho("\tpip3 --no-cache-dir install -U keep==" + str(version), fg='green')
                click.secho("\n\n")


def first_time_use(ctx):
    click.secho("Detected fresh installation. Initializing environment in ~/.keep directory", fg='green')
    for _ in range(2):
        click.echo('.', nl=False)
        time.sleep(0.5)
    click.echo('.OK', nl=True)

    os.mkdir(dir_path)

    # register()
    sys.exit(0)


def list_commands(ctx):
    table_data = [['Id', 'Command', 'Description', 'Alias']]
    no_of_columns = len(table_data[0])

    commands = read_commands()
    for i, (cmd, fields) in enumerate(commands.items()):
        table_data.append([str(i + 1), '$ ' + cmd, fields['desc'], fields['alias']])

    table = AsciiTable(table_data)
    max_width = table.table_width//3

    for i in range(len(table_data) - 1):
        for j in range(no_of_columns):
            data = table.table_data[i + 1][j]
            if len(data) > max_width:
                table.table_data[i + 1][j] = '\n'.join(wrap(data, max_width))

    table.inner_row_border = True
    print(table.table)


def log(ctx, message):
    """Prints log when verbose set to True."""
    if ctx.verbose:
        ctx.log(message)


# def push(ctx):
#     credentials_file = os.path.join(dir_path, '.credentials')
#     credentials = json.loads(open(credentials_file, 'r').read())
#     json_path = os.path.join(dir_path, 'commands.json')
#     credentials['commands'] = open(json_path).read()
#     url = api_url + '/push'
#     if click.confirm("This will overwrite the saved "
#                      "commands on the server. Proceed?", default=True):
#         r = requests.post(url, json=credentials)
#         if r.status_code == 200:
#             click.echo("Server successfully updated.")


# def pull(ctx, overwrite):
#     credentials_file = os.path.join(dir_path, '.credentials')
#     credentials = json.loads(open(credentials_file, 'r').read())
#     json_path = os.path.join(dir_path, 'commands.json')
#     url = api_url + '/pull'

#     r = requests.post(url, json=credentials)
#     if r.status_code == 200:
#         commands = json.loads(r.json()['commands'])

#     if not overwrite:
#         my_commands = read_commands()
#         commands.update(my_commands)

#     if not overwrite or (
#         overwrite and click.confirm(
#             "This will overwrite the locally saved commands. Proceed?", default=True)):
#             with open(json_path, 'w') as f:
#                 f.write(json.dumps(commands))
#             click.echo("Local database successfully updated.")


# def register():

#     if not os.path.exists(dir_path):
#         click.secho("\n[CRITICAL] {0} does not exits.\nPlease run 'keep init',"
#                     " and try registering again.\n".format(dir_path),
#                     fg="red", err=True)
#         sys.exit(1)

#     # User may not choose to register and work locally.
#     # Registration is required to push the commands to server
#     if click.confirm('Proceed to register?', abort=True, default=True):
#         # Verify for existing user
#         click.echo("Your credentials will be saved in the ~/.keep directory.")
#         email = click.prompt('Email', confirmation_prompt=True)
#         json_res = {'email': email}
#         click.echo('Verifying with existing users...')
#         r = requests.post('https://keep-cli.herokuapp.com/check-user', json=json_res)
#         if r.json()['exists']:
#             click.secho('User already exists !', fg='red')
#             email = click.prompt('Email', confirmation_prompt=True)
#             json_res = {'email': email}
#             r = requests.post('https://keep-cli.herokuapp.com/check-user', json=json_res)
#         # Generate password for the user
#         chars = string.ascii_letters + string.digits
#         password = ''.join(random.choice(chars) for _ in range(255))
#         credentials_file = os.path.join(dir_path, '.credentials')
#         credentials = {
#             'email': email,
#             'password': password
#         }
#         click.secho("Generated password for " + email, fg='cyan')
#         # Register over the server
#         click.echo("Registering new user ...")
#         json_res = {
#             'email': email,
#             'password': password
#         }
#         r = requests.post('https://keep-cli.herokuapp.com/register', json=json_res)
#         if r.status_code == 200:
#             click.secho("User successfully registered !", fg='green')
#             # Save the credentials into a file
#             with open(credentials_file, 'w+') as f:
#                 f.write(json.dumps(credentials))
#             click.secho(password, fg='cyan')
#             click.secho("Credentials file saved at ~/.keep/.credentials.json", fg='green')
#     sys.exit(0)


def remove_command(cmd):
    commands = read_commands()
    if cmd in commands:
        del commands[cmd]
        write_commands(commands)
    else:
        click.echo('Command - {} - does not exist.'.format(cmd))


def save_command(cmd, desc, alias=""):
    json_path = os.path.join(dir_path, 'commands.json')
    commands = {}
    if os.path.exists(json_path):
        commands = json.loads(open(json_path, 'r').read())
    fields = {'desc': desc, 'alias': alias}
    commands[cmd] = fields
    with open(json_path, 'w') as f:
        f.write(json.dumps(commands))


def read_commands():
    json_path = os.path.join(dir_path, 'commands.json')
    if not os.path.exists(json_path):
        return None
    commands = json.loads(open(json_path, 'r').read())
    return commands


def write_commands(commands):
    json_path = os.path.join(dir_path, 'commands.json')
    with open(json_path, 'w') as f:
        f.write(json.dumps(commands))


def grep_commands(pattern):
    commands = read_commands()
    result = None
    if commands:
        result = []
        for i, (cmd, fields) in enumerate(commands.items()):
            desc = fields['desc']
            alias = fields['alias']
            if pattern.isdigit() and pattern == str(i + 1):
                result.clear()
                result.append((cmd, desc))
                break
            else:
                if alias == pattern and alias.strip() != "":
                    result.clear()
                    result.append((cmd, desc))
                    break
                if re.search(pattern, cmd + " :: " + desc):
                    result.append((cmd, desc))
                    continue
                # Show if all the parts of the pattern are in one command/desc
                keywords_len = len(pattern.split())
                i_keyword = 0
                for keyword in pattern.split():
                    if keyword.lower() in cmd.lower() or keyword.lower() in desc.lower():
                        i_keyword += 1
                if i_keyword == keywords_len:
                    result.append((cmd, desc))
    return result


def select_command(commands):
    click.echo("", err=True)
    for idx, command in enumerate(commands):
        cmd, desc = command
        click.secho(f" {idx + 1} \t", nl=False, fg='yellow', err=True)
        click.secho(f" {cmd} :: {desc}", fg='green', err=True)

    selection = 1
    while True and len(commands) > 1:
        click.echo("", err=True)
        selection = click.prompt(
            f"Select a command [1-{len(commands)}] (0 to cancel)",
            type=int,
            err=True
        )
        if selection in range(len(commands) + 1):
            break
        click.echo("Number is not in range", err=True)
    return selection - 1


def edit_commands(commands, editor=None, edit_header=""):
    edit_msg = [edit_header]
    for cmd, fields in commands.items():
        desc, alias = fields['desc'], fields['alias']
        cmd = json.dumps(cmd)
        desc = json.dumps(desc)
        alias = json.dumps(alias)
        edit_msg.append("{} :: {} :: {}".format(cmd, alias, desc))
    edited = click.edit('\n'.join(edit_msg), editor=editor)

    command_regex = re.compile(r'(\".*\")\s*::\s*(\".*\")\s*::\s*(\".*\")')
    new_commands = {}
    if edited:
        for line in edited.split('\n'):
            if (line.startswith('#') or line == ""):
                continue
            re_match = command_regex.search(line)
            if re_match and len(re_match.groups()) == 3:
                cmd, alias, desc = re_match.groups()
                try:
                    cmd = json.loads(cmd)
                    desc = json.loads(desc)
                    alias = json.loads(alias)
                except ValueError:
                    click.echo("Error parsing json from edit file.")
                    return None
                new_commands[cmd] = {'desc': desc, 'alias': alias}
            else:
                click.echo("Could not read line '{}'".format(line))
    return new_commands


def format_commands(commands):
    res = []
    for cmd, fields in commands.items():
        desc, alias = fields['desc'], fields['alias']
        res.append("$ {} :: {} :: {}".format(cmd, alias, desc))
    return res


def create_pcmd(command):
    class KeepCommandTemplate(string.Template):
        default_sep = '='
        idpattern = r'[_a-z][_a-z0-9{}]*'.format(default_sep)

        def __init__(self, template):
            super().__init__(template)
    return KeepCommandTemplate(command)


def get_params_in_pcmd(pcmd):
    patt = pcmd.pattern
    params = []
    defaults = []
    raw = []
    for match in re.findall(patt, pcmd.template):
        var = match[1] or match[2]
        svar = var.split(pcmd.default_sep)
        p, d = svar[0], pcmd.default_sep.join(svar[1:])
        if p and p not in params:
            raw.append(var)
            params.append(p)
            defaults.append(d)
    return raw, params, defaults


def substitute_pcmd(pcmd, kargs, safe=False):
    if safe:
        return pcmd.safe_substitute(**kargs)
    else:
        return pcmd.substitute(**kargs)


def get_github_token():
    token_file_path = os.path.join(dir_path, '.credentials')
    token = None

    if os.path.exists(token_file_path):
        with open(token_file_path) as token_file:
            token = json.load(token_file)

    if not token or 'token' not in token or not token['token']:
        click.secho('Github access token not found :(', fg='red')
        click.secho('Use the "keep github_token" command to store the token first', fg='red')
        return None

    return token



import os
import click
from keep import cli, utils


@click.command('list', short_help='Shows the saved commands.')
@cli.pass_context
def cli(ctx):
    """Shows the saved commands."""
    json_path = os.path.join(os.path.expanduser('~'), '.keep', 'commands.json')
    if not os.path.exists(json_path):
        click.echo('No commands to show. Add one by `keep new`.')
    else:
        utils.list_commands(ctx)



import click
import json
from keep import cli, utils
from github import Github, InputFileContent


@click.command('push', short_help='Push commands to a secret GitHub gist.')
@cli.pass_context
def cli(ctx):
    """Push commands to a secret GitHub gist."""

    token = utils.get_github_token()
    if not token:
        return


    gist_url = f"https://gist.github.com/{token['gist']}"
    prompt_str = f"[CRITICAL] Overwrite upstream gist with local commands?\nGist URL : {gist_url}"
    click.confirm(prompt_str, abort=True)

    commands = utils.read_commands()
    if not commands:
        click.echo("No commands to push. Add one by 'keep new'. ")
    else:
        hub = Github(token['token'])
        gist = hub.get_gist(token['gist'])
        gist.edit(files={'commands.json': InputFileContent(json.dumps(commands))})
        click.echo("Done!")


import click
from keep import cli, utils

@click.command("new", short_help="Saves a new command.")
@click.option("--cmd", help="The command to save")
@click.option("--desc", help="The description of the command")
@click.option("--alias", default="", help="The alias of the command")
@cli.pass_context
def cli(ctx, cmd, desc, alias):
    """Saves a new command"""
    if not cmd:
        cmd = click.prompt("Command")
    if not desc:
        desc = click.prompt("Description")
    if not alias:
        alias = click.prompt("Alias (optional)", default="")
    utils.save_command(cmd, desc, alias)

    utils.log(ctx, "Saved the new command - {} - with the description - {}.".format(cmd, desc))



import os
import click
from keep import cli, utils


@click.command('push', short_help='Pushes the local database to remote.')
@cli.pass_context
def cli(ctx):
    """Pushes the local database to remote."""
    credentials_path = os.path.join(os.path.expanduser('~'), '.keep', '.credentials')
    if not os.path.exists(credentials_path):
        click.echo('You are not registered.')
        utils.register()
    else:
        utils.push(ctx)



import click
from keep import cli, utils
import os


@click.command('edit', short_help='Edit a saved command.')
@click.option('--editor', help='Editor to use')
@cli.pass_context
def cli(ctx, editor):
    """Edit saved commands."""
    commands = utils.read_commands()
    if commands is None:
        click.echo("No commands to edit, Add one by 'keep new'. ")
    else:
        edit_header = "# Unchanged file will abort the operation\n"
        new_commands = utils.edit_commands(commands, editor, edit_header)
        if new_commands and new_commands != commands:
            click.echo("Replace:\n")
            click.secho("\t{}".format('\n\t'.join(utils.format_commands(commands))),
                        fg="green")
            click.echo("With:\n\t")
            click.secho("\t{}".format('\n\t'.join(utils.format_commands(new_commands))),
                        fg="green")
            if click.confirm("", default=False):
                utils.write_commands(new_commands)
        elif new_commands == {}:
            dir_path = os.path.join(os.path.expanduser('~'), '.keep')
            json_path = os.path.join(dir_path, 'commands.json')
            if click.confirm('Delete all commands ?', abort=True):
                os.remove(json_path)



import click
import json
import os
from keep import cli
from github import Github, InputFileContent


@click.command('github_token', short_help='Register a GitHub Token to use GitHub Gists as a backup.')
@cli.pass_context
def cli(ctx):
    """Register a GitHub Token to use GitHub Gists as a backup."""

    dir_path = os.path.join(os.path.expanduser('~'), '.keep', '.credentials')
    existing_token = None
    if os.path.exists(dir_path):
        with open(dir_path) as f:
            existing_token = json.load(f)

        if click.confirm('[CRITICAL] Reset token saved in ~/.keep/.credentials ?',
                         abort=True):
            os.remove(dir_path)

    token = click.prompt('Create a GITHUB TOKEN with permission to Gists and paste here ')
    gist_id = existing_token.get('gist') if existing_token else None
    click.echo("\nAwesome! Your GitHub token has been registered. Now, we need a GIST ID to store the commands in keep.")
    prompt_str = 'If you already have a GitHub Gist created by keep, paste the GIST ID here, or else press Enter (we will create a Gist for you). '
    new_gist_id = click.prompt(prompt_str, default='', show_default=False)

    if new_gist_id:
        gist_id = new_gist_id

    if not gist_id:
        g = Github(token)
        gist = g.get_user().create_gist(False,
                                        files={'commands.json': InputFileContent('{}')},
                                        description='Backup for keep - https://github.com/OrkoHunter/keep')
        gist_id = gist.id

    with open(dir_path, 'w') as f:
        json.dump({'token': token, 'gist': gist_id}, f)

    click.echo("Done! Gist URL - https://gist.github.com/" + gist_id)
    click.echo("Your token and gist ID has been stored inside ~/.keep/.credentials")
    click.echo("You can now use `keep push` or `keep pull`")


import click
import os
from keep import cli, utils
from github import Github


@click.command('pull', short_help='Pull commands from saved GitHub gist.')
@cli.pass_context
def cli(ctx):
    """Pull commands from saved GitHub gist."""

    commands_file_path = os.path.join(utils.dir_path, 'commands.json')
    token = utils.get_github_token()
    if not token:
        return

    hub = Github(token['token'])
    gist = hub.get_gist(token['gist'])

    gist_url = f"https://gist.github.com/{token['gist']}"
    prompt_str = f"[CRITICAL] Replace local commands with GitHub gist\nGist URL : {gist_url} ?"
    if click.confirm(prompt_str, abort=True):
        pass

    """Using `w+` so it create the file if doesn't exist (Issue #64)"""
    with open(commands_file_path, 'w+') as commands_file:
        commands_file.write(gist.files['commands.json'].content)
    click.echo("Done!")



from shlex import quote

import click
from keep import cli, utils


@click.command('completion', short_help='Command completion helper.')
@click.option('--bash', 'shell', flag_value='bash', default=True)
@click.option('--zsh', 'shell', flag_value='zsh')
@cli.pass_context
def cli(ctx, shell):
    """Completion helpers for keep"""
    commands = utils.read_commands()

    if shell == 'zsh':
        for cmd, fields in commands.items():
            print("{cmd}:{desc}".format(cmd=cmd, desc=fields['desc']))
    elif shell == 'bash':
        for cmd in commands.keys():
            print(quote(cmd))



import os
import click
from keep import cli, utils


@click.command('pull', short_help='Updates the local database with remote.')
@click.option('--overwrite', is_flag=True, help='Overwrite local commands')
@cli.pass_context
def cli(ctx, overwrite):
    """Updates the local database with remote."""
    credentials_path = os.path.join(os.path.expanduser('~'), '.keep', '.credentials')
    if not os.path.exists(credentials_path):
        click.echo('You are not registered.')
        utils.register()
    else:
        utils.pull(ctx, overwrite)



import click
from keep import cli, utils, about

@click.command('update', short_help='Check for an update of Keep.')
@cli.pass_context
def cli(ctx):
    """Check for an update of Keep."""
    utils.check_update(ctx, forced=True)
    click.secho("Keep is at its latest version v{}".format(about.__version__), fg='green')



import os
import shutil
import click
from keep import cli, utils

@click.command('init', short_help='Initializes the CLI.')
@cli.pass_context
def cli(ctx):
    """Initializes the CLI environment."""
    dir_path = os.path.join(os.path.expanduser('~'), '.keep')
    if os.path.exists(dir_path):
        if click.confirm('[CRITICAL] Remove everything inside ~/.keep ?', abort=True):
            shutil.rmtree(dir_path)
    utils.first_time_use(ctx)



import os
import sys
import click

from keep import utils

CONTEXT_SETTINGS = dict(auto_envvar_prefix='KEEP')


class Context(object):

    def __init__(self):
        self.verbose = False
        self.home = os.getcwd()

    def log(self, msg, *args):
        """Logs a message to stderr."""
        if args:
            msg %= args
        click.echo(msg, file=sys.stderr)

    def vlog(self, msg, *args):
        """Logs a message to stderr only if verbose is enabled."""
        if self.verbose:
            self.log(msg, *args)


# Directory containing the commands modules
cmd_folder = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                          'commands'))
pass_context = click.make_pass_decorator(Context, ensure=True)


class KeepCLI(click.MultiCommand):

    def pre_process(self, ctx):
        dir_path = os.path.join(os.path.expanduser('~'), '.keep')
        if not os.path.exists(dir_path):
            utils.first_time_use(ctx)
        else:
            utils.check_update(ctx)

    def list_commands(self, ctx):
        self.pre_process(ctx)
        rv = []
        for filename in os.listdir(cmd_folder):
            if filename.endswith('.py') and \
               filename.startswith('cmd_'):
                rv.append(filename[4:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        self.pre_process(ctx)
        try:
            if sys.version_info[0] == 2:
                name = name.encode('ascii', 'replace')
            mod = __import__('keep.commands.cmd_' + name,
                             None, None, ['cli'])
        except ImportError:
            return
        return mod.cli


@click.command(cls=KeepCLI, context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', is_flag=True,
              help='Enables verbose mode.')
@pass_context
def cli(ctx, verbose):
    """
    Keep and view shell commands in terminal only.

    Read more at https://github.com/orkohunter/keep
    """
    ctx.verbose = verbose



import os
import click
from keep import cli, utils

@click.command('register', short_help='Register user over server.')
@cli.pass_context
def cli(ctx):
    """Register user over server."""
    dir_path = os.path.join(os.path.expanduser('~'), '.keep', '.credentials')
    if os.path.exists(dir_path):
        if click.confirm('[CRITICAL] Reset credentials saved in ~/.keep/.credentials ?', abort=True):
            os.remove(dir_path)
    utils.register()



import click
from keep import cli, utils


@click.command('rm', short_help='Deletes a saved command.')
@click.argument('pattern')
@cli.pass_context
def cli(ctx, pattern):
    """Deletes a saved command."""
    matches = utils.grep_commands(pattern)
    if matches:
        selected = utils.select_command(matches)
        if selected >= 0:
            cmd, desc = matches[selected]
            command = "$ {} :: {}".format(cmd, desc)
            if click.confirm("Remove\n\t{}\n\n?".format(command), default=True):
                utils.remove_command(cmd)
                click.echo('Command successfully removed!')
    elif matches == []:
        click.echo('No saved commands matches the pattern {}'.format(pattern))
    else:
        click.echo("No commands to remove, Add one by 'keep new'. ")



import os
import click
from keep import cli, utils


@click.command('run', short_help='Executes a saved command.',
               context_settings=dict(ignore_unknown_options=True))
@click.argument('pattern', required=False)
@click.argument('arguments', nargs=-1, type=click.UNPROCESSED)
@click.option('--safe', is_flag=True, help='Ignore missing arguments')
@click.option('-n', '--no-confirm', is_flag=True, help='Don\'t ask confirm before running')
@cli.pass_context
def cli(ctx, pattern, arguments, safe, no_confirm):
    """Executes a saved command."""

    if not pattern:
        pattern = "(.*?s)"
    matches = utils.grep_commands(pattern)
    if matches:
        selected = utils.select_command(matches)
        if selected >= 0:
            cmd, desc = matches[selected]
            pcmd = utils.create_pcmd(cmd)
            raw_params, params, defaults = utils.get_params_in_pcmd(pcmd)

            arguments = list(arguments)
            kargs = {}
            for r, p, d in zip(raw_params, params, defaults):
                if arguments:
                    val = arguments.pop(0)
                    click.echo(f"{p}: {val}", err=True)
                    kargs[r] = val
                elif safe:
                    if d:
                        kargs[r] = d
                else:
                    p_default = d if d else None
                    val = click.prompt(f"Enter value for '{p}'", default=p_default, err=True)
                    kargs[r] = val

            final_cmd = utils.substitute_pcmd(pcmd, kargs, safe)

            if no_confirm:
                isconfirmed = True
            else:
                command = f"$ {final_cmd} :: {desc}"
                isconfirmed = click.confirm(f"Execute\n\t{command}\n?", default=True, err=True)

            if isconfirmed:
                os.system(final_cmd)

    elif matches == []:
        click.echo(f'No saved commands matches the pattern {pattern}', err=True)
    else:
        click.echo("No commands to run, Add one by 'keep new'. ", err=True)



import click
from keep import cli, utils


@click.command('grep', short_help='Searches for a saved command.')
@click.argument('pattern')
@cli.pass_context
def cli(ctx, pattern):
    """Searches for a saved command."""
    matches = utils.grep_commands(pattern)
    if matches:
        for cmd, desc in matches:
            click.secho("$ {} :: {}".format(cmd, desc), fg='green')
    elif matches == []:
        click.echo('No saved commands matches the pattern {}'.format(pattern))
    else:
        click.echo('No commands to show. Add one by `keep new`.')
